@@ -1,4 +1,4 @@
-# Copyright (c) 2016 Iotic Labs Ltd. All rights reserved.
+# Copyright (c) 2019 Iotic Labs Ltd. All rights reserved.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -15,25 +15,29 @@
 
 from sys import version_info, getrecursionlimit
 from functools import partial
-from io import BytesIO
-from unittest import TestCase
+from io import BytesIO, SEEK_END
+from unittest import TestCase, skipUnless
 from pprint import pformat
 from decimal import Decimal
 from struct import pack
 from collections import OrderedDict
 
 from ubjson import (dump as ubjdump, dumpb as ubjdumpb, load as ubjload, loadb as ubjloadb, EncoderException,
-                    DecoderException)
+                    DecoderException, EXTENSION_ENABLED)
 from ubjson.markers import (TYPE_NULL, TYPE_NOOP, TYPE_BOOL_TRUE, TYPE_BOOL_FALSE, TYPE_INT8, TYPE_UINT8, TYPE_INT16,
                             TYPE_INT32, TYPE_INT64, TYPE_FLOAT32, TYPE_FLOAT64, TYPE_HIGH_PREC, TYPE_CHAR, TYPE_STRING,
                             OBJECT_START, OBJECT_END, ARRAY_START, ARRAY_END, CONTAINER_TYPE, CONTAINER_COUNT)
+from ubjson.compat import INTEGER_TYPES
+# Pure Python versions
+from ubjson.encoder import dump as ubjpuredump, dumpb as ubjpuredumpb
+from ubjson.decoder import load as ubjpureload, loadb as ubjpureloadb
 
 PY2 = version_info[0] < 3
 
 if PY2:  # pragma: no cover
     def u(obj):
         """Casts obj to unicode string, unless already one"""
-        return obj if isinstance(obj, unicode) else unicode(obj)  # noqa  pylint: disable=undefined-variable
+        return obj if isinstance(obj, unicode) else unicode(obj)  # noqa: F821 pylint: disable=undefined-variable
 else:  # pragma: no cover
     def u(obj):
         """Casts obj to unicode string, unless already one"""
@@ -44,11 +48,11 @@ class TestEncodeDecodePlain(TestCase):  # pylint: disable=too-many-public-method
 
     @staticmethod
     def ubjloadb(raw, *args, **kwargs):
-        return ubjloadb(raw, *args, **kwargs)
+        return ubjpureloadb(raw, *args, **kwargs)
 
     @staticmethod
     def ubjdumpb(obj, *args, **kwargs):
-        return ubjdumpb(obj, *args, **kwargs)
+        return ubjpuredumpb(obj, *args, **kwargs)
 
     @staticmethod
     def __format_in_out(obj, encoded):
@@ -102,15 +106,17 @@ class TestEncodeDecodePlain(TestCase):  # pylint: disable=too-many-public-method
             self.ubjloadb(b'')
 
     def test_invalid_data(self):
-        with self.assertRaises(TypeError):
-            self.ubjloadb(123)
+        for invalid in (u('unicode'), 123):
+            with self.assertRaises(TypeError):
+                self.ubjloadb(invalid)
 
     def test_trailing_input(self):
         self.assertEqual(self.ubjloadb(TYPE_BOOL_TRUE * 10), True)
 
     def test_invalid_marker(self):
-        with self.assertRaises(DecoderException):
+        with self.assertRaises(DecoderException) as ctx:
             self.ubjloadb(b'A')
+        self.assertTrue(isinstance(ctx.exception.position, INTEGER_TYPES + (type(None),)))
 
     def test_bool(self):
         self.assertEqual(self.ubjdumpb(True), TYPE_BOOL_TRUE)
@@ -250,19 +256,22 @@ class TestEncodeDecodePlain(TestCase):  # pylint: disable=too-many-public-method
         # insufficient length
         with self.assertRaises(DecoderException):
             self.ubjloadb(ARRAY_START + CONTAINER_TYPE + TYPE_UINT8 + CONTAINER_COUNT + TYPE_UINT8 + b'\x02' + b'\x01')
-        self.check_enc_dec(b'')
-        self.check_enc_dec(b'\x01' * 4)
-        self.assertEqual(self.ubjloadb(self.ubjdumpb(b'\x04' * 4), no_bytes=True), [4] * 4)
-        self.check_enc_dec(b'largebinary' * 100)
+        for cast in (bytes, bytearray):
+            self.check_enc_dec(cast(b''))
+            self.check_enc_dec(cast(b'\x01' * 4))
+            self.assertEqual(self.ubjloadb(self.ubjdumpb(cast(b'\x04' * 4)), no_bytes=True), [4] * 4)
+            self.check_enc_dec(cast(b'largebinary' * 100))
 
     def test_array_fixed(self):
         raw_start = ARRAY_START + CONTAINER_TYPE + TYPE_INT8 + CONTAINER_COUNT + TYPE_UINT8
         self.assertEqual(self.ubjloadb(raw_start + b'\x00'), [])
 
-        # fixed-type + count
-        self.assertEqual(self.ubjloadb(ARRAY_START + CONTAINER_TYPE + TYPE_NULL + CONTAINER_COUNT + TYPE_UINT8 +
-                                       b'\x05'),
-                         [None] * 5)
+        # fixed types + count
+        for ubj_type, py_obj in ((TYPE_NULL, None), (TYPE_BOOL_TRUE, True), (TYPE_BOOL_FALSE, False)):
+            self.assertEqual(
+                self.ubjloadb(ARRAY_START + CONTAINER_TYPE + ubj_type + CONTAINER_COUNT + TYPE_UINT8 + b'\x05'),
+                [py_obj] * 5
+            )
         self.assertEqual(self.ubjloadb(raw_start + b'\x03' + (b'\x01' * 3)), [1, 1, 1])
 
         # invalid type
@@ -297,6 +306,34 @@ class TestEncodeDecodePlain(TestCase):  # pylint: disable=too-many-public-method
                                        TYPE_NOOP +
                                        TYPE_UINT8 + b'\x01'), [1])
 
+    def test_object_invalid(self):
+        # negative length
+        with self.assertRaises(DecoderException):
+            self.ubjloadb(OBJECT_START + CONTAINER_COUNT + self.ubjdumpb(-1))
+
+        with self.assertRaises(EncoderException):
+            self.ubjdumpb({123: 'non-string key'})
+
+        with self.assertRaises(EncoderException):
+            self.ubjdumpb({'fish': type(list)})
+
+        # invalid key size type
+        with self.assertRaises(DecoderException):
+            self.ubjloadb(OBJECT_START + TYPE_NULL)
+
+        # invalid key size, key too short, key invalid utf-8, no value
+        for suffix in (b'\x81', b'\x01', b'\x01' + b'\xfe', b'\x0101'):
+            with self.assertRaises(DecoderException):
+                self.ubjloadb(OBJECT_START + TYPE_INT8 + suffix)
+
+        # invalid items() method
+        class BadDict(dict):
+            def items(self):
+                return super(BadDict, self).keys()
+
+        with self.assertRaises(ValueError):
+            self.ubjdumpb(BadDict({'a': 1, 'b': 2}))
+
     def test_object(self):
         # custom hook
         with self.assertRaises(TypeError):
@@ -305,29 +342,15 @@ class TestEncodeDecodePlain(TestCase):  # pylint: disable=too-many-public-method
         self.ubjloadb(self.ubjdumpb({}), object_pairs_hook=None)
 
         for hook in (None, OrderedDict):
-            loadb = partial(ubjloadb, object_pairs_hook=hook)
+            check_enc_dec = partial(self.check_enc_dec, object_pairs_hook=hook)
 
             self.assertEqual(self.ubjdumpb({}), OBJECT_START + OBJECT_END)
             self.assertEqual(self.ubjdumpb({'a': None}, container_count=True),
                              (OBJECT_START + CONTAINER_COUNT + TYPE_UINT8 + b'\x01' + TYPE_UINT8 + b'\x01' +
                               'a'.encode('utf-8') + TYPE_NULL))
-            self.check_enc_dec({})
-            # negative length
-            with self.assertRaises(DecoderException):
-                loadb(OBJECT_START + CONTAINER_COUNT + self.ubjdumpb(-1))
-            with self.assertRaises(EncoderException):
-                self.ubjdumpb({123: 'non-string key'})
-            with self.assertRaises(EncoderException):
-                self.ubjdumpb({'fish': type(list)})
-            # invalid key size type
-            with self.assertRaises(DecoderException):
-                loadb(OBJECT_START + TYPE_NULL)
-            # invalid key size, key too short, key invalid utf-8, no value
-            for suffix in (b'\x81', b'\x01', b'\x01' + b'\xfe', b'\x0101'):
-                with self.assertRaises(DecoderException):
-                    loadb(OBJECT_START + TYPE_INT8 + suffix)
-            self.check_enc_dec({'longkey1' * 65: 1})
-            self.check_enc_dec({'longkey2' * 4096: 1})
+            check_enc_dec({})
+            check_enc_dec({'longkey1' * 65: 1})
+            check_enc_dec({'longkey2' * 4096: 1})
 
             obj = {'int': 123,
                    'longint': 9223372036854775807,
@@ -345,7 +368,7 @@ class TestEncodeDecodePlain(TestCase):  # pylint: disable=too-many-public-method
                    'bytes_array': b'1234',
                    'object': {'another one': 456, 'yet another': {'abc': True}}}
             for opts in ({'container_count': False}, {'container_count': True}):
-                self.check_enc_dec(obj, **opts)
+                check_enc_dec(obj, **opts)
 
         # dictionary key sorting
         obj1 = OrderedDict.fromkeys('abcdefghijkl')
@@ -359,7 +382,7 @@ class TestEncodeDecodePlain(TestCase):  # pylint: disable=too-many-public-method
         raw_start = OBJECT_START + CONTAINER_TYPE + TYPE_INT8 + CONTAINER_COUNT + TYPE_UINT8
 
         for hook in (None, OrderedDict):
-            loadb = partial(ubjloadb, object_pairs_hook=hook)
+            loadb = partial(self.ubjloadb, object_pairs_hook=hook)
 
             self.assertEqual(loadb(raw_start + b'\x00'), {})
             self.assertEqual(loadb(raw_start + b'\x03' + (TYPE_UINT8 + b'\x02' + b'aa' + b'\x01' +
@@ -404,7 +427,7 @@ class TestEncodeDecodePlain(TestCase):  # pylint: disable=too-many-public-method
             if PY2:  # pragma: no cover
                 # interning of unicode not supported
                 self.assertEqual(key1, key2)
-            else:
+            else:  # pragma: no cover
                 self.assertIs(key1, key2)
 
     def test_circular(self):
@@ -470,6 +493,9 @@ class TestEncodeDecodePlain(TestCase):  # pylint: disable=too-many-public-method
 
         with self.assert_raises_regex(EncoderException, 'Cannot encode item'):
             self.ubjdumpb(obj1)
+        # explicit None should behave the same as no default
+        with self.assert_raises_regex(EncoderException, 'Cannot encode item'):
+            self.ubjdumpb(obj1, default=None)
 
         with self.assert_raises_regex(EncoderException, '__test__marker__'):
             dumpb_default(self)
@@ -493,28 +519,54 @@ class TestEncodeDecodePlain(TestCase):  # pylint: disable=too-many-public-method
 
         self.check_enc_dec({'a': 1, 'b': {2, 3, 4}}, object_hook=object_hook, default=default)
 
+        class UnHandled(object):
+            pass
+
+        with self.assertRaises(EncoderException):
+            self.check_enc_dec({'a': 1, 'b': UnHandled()}, object_hook=object_hook, default=default)
+
+
+@skipUnless(EXTENSION_ENABLED, 'Extension not enabled')
+class TestEncodeDecodePlainExt(TestEncodeDecodePlain):
+
+    @staticmethod
+    def ubjloadb(raw, *args, **kwargs):
+        return ubjloadb(raw, *args, **kwargs)
+
+    @staticmethod
+    def ubjdumpb(obj, *args, **kwargs):
+        return ubjdumpb(obj, *args, **kwargs)
+
 
 class TestEncodeDecodeFp(TestEncodeDecodePlain):
     """Performs tests via file-like objects (BytesIO) instead of bytes instances"""
 
     @staticmethod
     def ubjloadb(raw, *args, **kwargs):
-        try:
-            raw = BytesIO(raw)
-        except TypeError:  # pylint: disable=bare-except
-            # Invalid raw input testing
-            raise
-        return ubjload(raw, *args, **kwargs)
+        return ubjpureload(BytesIO(raw), *args, **kwargs)
 
     @staticmethod
     def ubjdumpb(obj, *args, **kwargs):
         out = BytesIO()
-        ubjdump(obj, out, *args, **kwargs)
+        ubjpuredump(obj, out, *args, **kwargs)
         return out.getvalue()
 
+    @staticmethod
+    def ubjload(fp, *args, **kwargs):
+        return ubjpureload(fp, *args, **kwargs)
+
+    @staticmethod
+    def ubjdump(obj, fp, *args, **kwargs):
+        return ubjpuredump(obj, fp, *args, **kwargs)
+
+    def test_decode_exception_position(self):
+        with self.assertRaises(DecoderException) as ctx:
+            self.ubjloadb(TYPE_STRING + TYPE_INT8 + b'\x01' + b'\xfe' + b'c0fefe' * 4)
+        self.assertEqual(ctx.exception.position, 4)
+
     def test_invalid_fp_dump(self):
         with self.assertRaises(AttributeError):
-            ubjdump(None, 1)
+            self.ubjdump(None, 1)
 
         class Dummy(object):
             write = 1
@@ -525,14 +577,14 @@ class TestEncodeDecodeFp(TestEncodeDecodePlain):
                 raise ValueError('invalid - %s' % repr(raw))
 
         with self.assertRaises(TypeError):
-            ubjdump(b'', Dummy)
+            self.ubjdump(b'', Dummy)
 
         with self.assertRaises(ValueError):
-            ubjdump(b'', Dummy2)
+            self.ubjdump(b'', Dummy2)
 
     def test_invalid_fp_load(self):
         with self.assertRaises(AttributeError):
-            ubjload(1)
+            self.ubjload(1)
 
         class Dummy(object):
             read = 1
@@ -544,31 +596,121 @@ class TestEncodeDecodeFp(TestEncodeDecodePlain):
                 raise ValueError('invalid - %d' % length)
 
         with self.assertRaises(TypeError):
-            ubjload(Dummy)
+            self.ubjload(Dummy)
 
         with self.assertRaises(ValueError):
-            ubjload(Dummy2)
+            self.ubjload(Dummy2)
 
     def test_fp(self):
-        obj = {"a": 123, "b": 456}
+        obj = {'a': 123, 'b': 456}
         output = BytesIO()
-        ubjdump({"a": 123, "b": 456}, output)
+        self.ubjdump(obj, output)
         output.seek(0)
-        self.assertEqual(ubjload(output), obj)
+        self.assertEqual(self.ubjload(output), obj)
 
-        # items which fit into extension decoder-internal read buffer (BUFFER_FP_SIZE in decoder.c)
+
+@skipUnless(EXTENSION_ENABLED, 'Extension not enabled')
+class TestEncodeDecodeFpExt(TestEncodeDecodeFp):
+
+    @staticmethod
+    def ubjloadb(raw, *args, **kwargs):
+        return ubjload(BytesIO(raw), *args, **kwargs)
+
+    @staticmethod
+    def ubjdumpb(obj, *args, **kwargs):
+        out = BytesIO()
+        ubjdump(obj, out, *args, **kwargs)
+        return out.getvalue()
+
+    @staticmethod
+    def ubjload(fp, *args, **kwargs):
+        return ubjload(fp, *args, **kwargs)
+
+    @staticmethod
+    def ubjdump(obj, fp, *args, **kwargs):
+        return ubjdump(obj, fp, *args, **kwargs)
+
+    # Seekable file-like object buffering
+    def test_fp_buffer(self):
+        output = BytesIO()
+
+        # items which fit into extension decoder-internal read buffer (BUFFER_FP_SIZE in decoder.c, extension only)
         obj2 = ['fishy' * 64] * 10
-        output = BytesIO()
-        ubjdump(obj, output)
         output.seek(0)
-        self.assertEqual(ubjload(output), obj)
+        self.ubjdump(obj2, output)
+        output.seek(0)
+        self.assertEqual(self.ubjload(output), obj2)
 
-        # larger than extension read buffer
-        obj2 = ['fishy' * 512] * 10
+        # larger than extension read buffer (extension only)
+        obj3 = ['fishy' * 512] * 10
         output.seek(0)
-        ubjdump(obj2, output)
+        self.ubjdump(obj3, output)
         output.seek(0)
-        self.assertEqual(ubjload(output), obj2)
+        self.assertEqual(self.ubjload(output), obj3)
+
+    # Multiple documents in same stream (issue #9)
+    def test_fp_multi(self):
+        obj = {'a': 123, 'b': b'some raw content'}
+        output = BytesIO()
+        count = 10
+
+        # Seekable an non-seekable runs
+        for _ in range(2):
+            output.seek(0)
+
+            for i in range(count):
+                obj['c'] = i
+                self.ubjdump(obj, output)
+
+            output.seek(0)
+            for i in range(count):
+                obj['c'] = i
+                self.assertEqual(self.ubjload(output), obj)
+
+            output.seekable = lambda: False
+
+    # Whole "token" in decoder input unavailable (in non-seekable file-like object)
+    def test_fp_callable_incomplete(self):
+        obj = [123, b'something']
+        # remove whole of last token (binary data 'something', without its length)
+        output = BytesIO(self.ubjdumpb(obj)[:-(len(obj[1]) + 1)])
+        output.seekable = lambda: False
+
+        with self.assert_raises_regex(DecoderException, 'Insufficient input'):
+            self.ubjload(output)
+
+    def test_fp_seek_invalid(self):
+        output = BytesIO()
+        self.ubjdump({'a': 333, 'b': 444}, output)
+        # pad with data (non-ubjson) to ensure buffering too much data
+        output.write(b' ' * 16)
+        output.seek(0)
+
+        output.seek_org = output.seek
+
+        # seek fails
+        def bad_seek(*_):
+            raise OSError('bad seek')
+
+        output.seek = bad_seek
+        with self.assert_raises_regex(OSError, 'bad seek'):
+            self.ubjload(output)
+
+        # decoding (lack of input) and seek fail - should get decoding failure
+        output.seek_org(0, SEEK_END)
+        with self.assert_raises_regex(DecoderException, 'Insufficient input'):
+            self.ubjload(output)
+
+        # seek is not callable
+        output.seek_org(0)
+        output.seek = True
+        with self.assert_raises_regex(TypeError, 'not callable'):
+            self.ubjload(output)
+
+        # decoding (lack of input) and seek not callable - should get decoding failure
+        output.seek_org(0, SEEK_END)
+        with self.assert_raises_regex(DecoderException, 'Insufficient input'):
+            self.ubjload(output)
 
 
 # def pympler_run(iterations=20):
