@@ -1,4 +1,4 @@
-# Copyright (c) 2016 Iotic Labs Ltd. All rights reserved.
+# Copyright (c) 2019 Iotic Labs Ltd. All rights reserved.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -26,10 +26,17 @@ decoded = ubjson.loadb(encoded)
 To use a file-like object as input/output, use dump() & load() methods instead.
 """
 
-from .encoder import dump, dumpb, EncoderException  # noqa
-from .decoder import load, loadb, DecoderException  # noqa
-from .compat import EXTENSION_ENABLED
+try:
+    from _ubjson import dump, dumpb, load, loadb
+    EXTENSION_ENABLED = True
+except ImportError:  # pragma: no cover
+    from .encoder import dump, dumpb
+    from .decoder import load, loadb
+    EXTENSION_ENABLED = False
 
-__version__ = '0.8.5'
+from .encoder import EncoderException
+from .decoder import DecoderException
+
+__version__ = '0.14.0'
 
 __all__ = ('EXTENSION_ENABLED', 'dump', 'dumpb', 'EncoderException', 'load', 'loadb', 'DecoderException')
